// Win32UI.cpp : Defines the entry point for the application.
//

#include "stdafx.h"
#include "Win32UI.h"

#define MAX_LOADSTRING 100

void Test1(HWND hWnd);
void Test2(HWND hWnd);

LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);

int APIENTRY _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	UNREFERENCED_PARAMETER(hPrevInstance);
	UNREFERENCED_PARAMETER(lpCmdLine);

	// Also tried COINIT_APARTMENTTHREADED
	CoInitializeEx(0, COINIT_MULTITHREADED);

	{
		WNDCLASSEX wcex = {0};

		wcex.cbSize = sizeof(WNDCLASSEX);

		wcex.style			= CS_HREDRAW | CS_VREDRAW;
		wcex.lpfnWndProc	= WndProc;
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= hInstance;
		wcex.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_WIN32UI));
		wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
		wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
		wcex.lpszMenuName	= MAKEINTRESOURCE(IDC_WIN32UI);
		wcex.lpszClassName	= L"Win32UI_Test";
		wcex.hIconSm		= LoadIcon(wcex.hInstance, MAKEINTRESOURCE(IDI_SMALL));

		if (RegisterClassEx(&wcex) == 0)
		{
			return 0;
		}

		HWND hWnd = CreateWindow(L"Win32UI_Test", L"Win32UI Test", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

		if (hWnd == 0)
		{
			return 0;
		}

		ShowWindow(hWnd, nCmdShow);
		UpdateWindow(hWnd);
	}

	MSG msg;

	// Main message loop:
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	CoUninitialize();

	return (int) msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;

	switch (message)
	{
	case WM_COMMAND:
		wmId    = LOWORD(wParam);
		wmEvent = HIWORD(wParam);
		// Parse the menu selections:
		switch (wmId)
		{
		case IDM_TEST1:
			Test1(hWnd);
			break;

		case IDM_TEST2:
			Test2(hWnd);
			break;

		case IDM_EXIT:
			DestroyWindow(hWnd);
			break;

		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
		break;

	case WM_PAINT:
		{
			PAINTSTRUCT ps = {0};
			HDC hdc = BeginPaint(hWnd, &ps);

			RECT r;
			GetClientRect(hWnd, &r);

			std::wstring strMsg = L"Use the Test menu to start a test";

			DrawText(hdc, strMsg.c_str(), -1, &r, 0);

			EndPaint(hWnd, &ps);
		}
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

// ********************************************************************
// Utilities
// ********************************************************************

// delete[] the result.
// Returns NULL on failure.
TCHAR *VStringAllocAndFormat(const TCHAR *format, va_list args)
{
	size_t bufferSize = (_tcslen(format) * 3) / 2;

	if (bufferSize < 64)
	{
		bufferSize = 64;
	}

	TCHAR *szResult = new(std::nothrow) TCHAR[bufferSize];

	while(true)
	{
		if (NULL == szResult)
		{
			::SetLastError(ERROR_NOT_ENOUGH_MEMORY);
			break;
		}

		int formatRes = _vsntprintf_s(szResult, bufferSize, _TRUNCATE, format, args);

		// formatRes negative on error; (bufferSize == formatRes) should be redundant for _vsntprintf.

		if ((0 > formatRes) || (bufferSize == formatRes))
		{
			delete [] szResult;
			bufferSize *= 2;
			szResult = new(std::nothrow) TCHAR[bufferSize];
		}
		else
		{
			break;
		}
	}

	return(szResult);
}

// Returns boolean success.
bool StringFormatToString(std::basic_string< TCHAR > *pResult, const TCHAR *format, ...)
{
	va_list args;
	va_start(args, format);
	TCHAR *result = VStringAllocAndFormat(format, args);
	va_end(args);

	if (result == NULL)
	{
		pResult->clear();
		return false;
	}

	*pResult = result;
	delete[] result;

	return true;
}

// Never pass AutoBuffer objects across process/DLL boundaries.
// (It's all inline so the new/delete calls may hit different heaps.)
template <typename T> class AutoBuffer
{
private:
	BYTE *m_pBuffer;
	size_t m_size;
private:
	AutoBuffer(const AutoBuffer &rhs); // disallow
	AutoBuffer &operator=(const AutoBuffer &rhs); // disallow
public:
	AutoBuffer() : m_pBuffer(0) , m_size(0) { }
	/* not virtual */ ~AutoBuffer() { Free(); } // Must not change Win32 LastError
	void Free() { delete[] m_pBuffer; m_pBuffer = 0; m_size = 0; } 		// Must not change Win32 LastError

	// Allocating a zero-length buffer is legal and results in a non-null buffer.
	// Make no assumptions about the content of the buffer after allocation.
	bool AllocateBytes(size_t s)
	{
		if (s == m_size && m_pBuffer != 0) { return true; }
		Free();
		m_pBuffer = new(std::nothrow) BYTE[s];
		if (m_pBuffer == 0) { m_size = 0; ::SetLastError(ERROR_NOT_ENOUGH_MEMORY); return false; }
		m_size = s;
		return true;
	}

	// If there's already a buffer big enough (or bigger) then do nothing, else (re-)allocate.
	bool AllocateMinimumBytes(size_t s)
	{
		if (s <= m_size && m_pBuffer != 0) { return true; }
		return AllocateBytes(s);
	}

	bool        AllocateElements(size_t c) {        return AllocateBytes(c * sizeof(T)); }
	bool AllocateMinimumElements(size_t c) { return AllocateMinimumBytes(c * sizeof(T)); }

	T *    GetBuffer()             { return reinterpret_cast<       T * >(m_pBuffer); }
	T *    GetBuffer()       const { return reinterpret_cast< const T * >(m_pBuffer); }
	size_t GetSizeBytes()    const { return m_size; }
	size_t GetSizeElements() const { return (m_size / sizeof(T)); }
	bool   IsAllocated()     const { return (m_pBuffer != 0); }
};

// ********************************************************************
// Functions to test pixel data.
// ********************************************************************

bool testImage(HWND hWnd, const int W, const int H, HDC hdc, RGBQUAD *rgbInput, const wchar_t *szTestName)
{
	for (int y = 0; y < H; ++y)
	{
		for (int x = 0; x < W; ++x)
		{
			COLORREF cr = GetPixel(hdc, x, y);
			RGBQUAD rgb = rgbInput[ y * W + x ];

			if (GetBValue(cr)   != rgb.rgbBlue
			||	GetGValue(cr)   != rgb.rgbGreen
			||	GetRValue(cr)   != rgb.rgbRed
			||	((cr>>24)%0xff) != rgb.rgbReserved)
			{
				std::wstring strMsg;
				StringFormatToString(&strMsg, L"MISMATCH %s @ %d,%d\n\n%d,%d,%d,%d\n%d,%d,%d,%d",
					szTestName, x, y,
					(int)GetBValue(cr), (int)GetGValue(cr), (int)GetRValue(cr), (int)((cr>>24)%0xff),
					(int)rgb.rgbBlue,   (int)rgb.rgbGreen,  (int)rgb.rgbRed,    (int)rgb.rgbReserved);

				MessageBox(hWnd, strMsg.c_str(), L"Win32UI", MB_OK);

				return false;
			}
		}
	}

	std::wstring strMsg = szTestName;
	strMsg += L" good.";
	MessageBox(hWnd, strMsg.c_str(), L"Win32UI", MB_OK);

	return true;
}

bool testPalette(HWND hWnd, const int W, const int H, const int ColCount, RGBQUAD *rgbInput, RGBQUAD * rgbPalette)
{
	for (int y = 0; y < H; ++y)
	{
		for (int x = 0; x < W; ++x)
		{
			RGBQUAD rgb = rgbInput[ y * W + x ];

			bool bFound = false;

			for (int c = 0; c < ColCount; ++c)
			{
				if (0 == memcmp(rgbPalette + c, &rgb, sizeof(RGBQUAD)))
				{
					bFound = true;
					break;
				}
			}

			if (!bFound)
			{
				std::wstring strMsg;
				StringFormatToString(&strMsg, L"NOT IN PALETTE @ %d,%d\n\n%d,%d,%d,%d",
					x, y,
					(int)rgb.rgbBlue,   (int)rgb.rgbGreen,  (int)rgb.rgbRed,    (int)rgb.rgbReserved);

				MessageBox(hWnd, strMsg.c_str(), L"Win32UI", MB_OK);

				return false;
			}
		}
	}

	MessageBox(hWnd, L"Palette good.", L"Win32UI", MB_OK);

	return true;
}

void reportDibColorTable(HWND hWnd, HDC hdc, int ColCount)
{
	RGBQUAD rgbDT[256];
	ZeroMemory(rgbDT, sizeof(rgbDT));
	GetDIBColorTable(hdc, 0, ColCount, rgbDT);

	for (int i = 0; i < 3; ++i)
	{
		RGBQUAD rgb = rgbDT[i];

		std::wstring strMsg;
		StringFormatToString(&strMsg, L"DIB Color Table @ %d\n\n%d,%d,%d,%d",
			i,
			(int)rgb.rgbBlue,   (int)rgb.rgbGreen,  (int)rgb.rgbRed,    (int)rgb.rgbReserved);

		MessageBox(hWnd, strMsg.c_str(), L"Win32UI", MB_OK);
	}

	PALETTEENTRY peSys[256];
	ZeroMemory(peSys, sizeof(peSys));
	GetSystemPaletteEntries(hdc, 0, 256, peSys);

	std::wstring strMsg;
	std::wstring strMsgAll = L"System Color Table:";

	for (int i = 0; i < 256; ++i)
	{
		if (i%4 == 0)
			strMsgAll += L"\n";

		PALETTEENTRY pe = peSys[i];

		StringFormatToString(&strMsg, L"[%d: %d,%d,%d F=%d] ",
			i, (int)pe.peBlue,   (int)pe.peGreen,  (int)pe.peRed,    (int)pe.peFlags);
		strMsgAll += strMsg;
	}

	MessageBox(hWnd, strMsgAll.c_str(), L"Win32UI", MB_OK);
}

// ********************************************************************
// Main test routine(s)
// ********************************************************************

void Test1(HWND hWnd)
{
	// 13x51 32-bit RGB image.
	RGBQUAD rgbInput[663] =
	{
		{ 140, 116, 81, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 143, 142, 142, 0 }, { 196, 190, 190, 0 }, { 202, 195, 195, 0 }, { 207, 200, 200, 0 }, { 213, 206, 206, 0 }, { 217, 209, 209, 0 }, { 221, 214, 214, 0 }, { 225, 217, 217, 0 }, { 228, 220, 220, 0 }, { 229, 221, 221, 0 }, { 231, 223, 223, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 143, 142, 140, 0 }, { 187, 181, 181, 0 }, { 194, 187, 187, 0 }, { 201, 194, 194, 0 }, { 207, 200, 200, 0 }, { 213, 206, 206, 0 }, { 218, 211, 211, 0 }, { 222, 214, 214, 0 }, { 227, 219, 219, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 143, 141, 139, 0 }, { 177, 171, 171, 0 }, { 185, 179, 179, 0 }, { 194, 187, 187, 0 }, { 202, 195, 195, 0 }, { 209, 202, 202, 0 }, { 215, 207, 207, 0 }, { 220, 213, 213, 0 }, { 225, 217, 217, 0 }, { 228, 221, 221, 0 }, { 230, 222, 222, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 140, 138, 0 }, { 168, 162, 162, 0 }, { 177, 171, 171, 0 }, { 187, 181, 181, 0 }, { 196, 190, 190, 0 }, { 205, 198, 198, 0 }, { 212, 205, 205, 0 }, { 218, 211, 211, 0 }, { 223, 215, 215, 0 }, { 227, 219, 219, 0 }, { 230, 222, 222, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 140, 136, 0 }, { 160, 154, 154, 0 }, { 170, 164, 164, 0 }, { 182, 176, 176, 0 }, { 192, 185, 185, 0 }, { 201, 194, 194, 0 }, { 209, 202, 202, 0 }, { 216, 208, 208, 0 }, { 222, 214, 214, 0 }, { 227, 219, 219, 0 }, { 229, 221, 221, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 139, 135, 0 }, { 153, 147, 147, 0 }, { 164, 159, 159, 0 }, { 176, 170, 170, 0 }, { 187, 181, 181, 0 }, { 198, 191, 191, 0 }, { 207, 200, 200, 0 }, { 215, 207, 207, 0 }, { 220, 213, 213, 0 }, { 226, 218, 218, 0 }, { 229, 221, 221, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 139, 133, 0 }, { 147, 142, 142, 0 }, { 159, 154, 154, 0 }, { 173, 167, 167, 0 }, { 185, 178, 178, 0 }, { 196, 189, 189, 0 }, { 205, 198, 198, 0 }, { 213, 206, 206, 0 }, { 219, 212, 212, 0 }, { 225, 217, 217, 0 }, { 228, 221, 221, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 138, 132, 0 }, { 142, 137, 137, 0 }, { 155, 150, 150, 0 }, { 169, 163, 163, 0 }, { 182, 176, 176, 0 }, { 194, 187, 187, 0 }, { 203, 196, 196, 0 }, { 212, 205, 205, 0 }, { 218, 211, 211, 0 }, { 225, 217, 217, 0 }, { 228, 221, 221, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 138, 131, 0 }, { 138, 133, 133, 0 }, { 151, 146, 146, 0 }, { 166, 161, 161, 0 }, { 180, 174, 174, 0 }, { 192, 185, 185, 0 }, { 202, 195, 195, 0 }, { 211, 204, 204, 0 }, { 218, 211, 211, 0 }, { 225, 217, 217, 0 }, { 228, 221, 221, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 137, 129, 0 }, { 135, 131, 131, 0 }, { 149, 144, 144, 0 }, { 164, 159, 159, 0 }, { 178, 172, 172, 0 }, { 190, 184, 184, 0 }, { 201, 194, 194, 0 }, { 210, 203, 203, 0 }, { 217, 210, 210, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 136, 128, 0 }, { 132, 128, 128, 0 }, { 147, 142, 142, 0 }, { 163, 157, 157, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 200, 193, 193, 0 }, { 209, 202, 202, 0 }, { 217, 210, 210, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 135, 127, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 135, 125, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 134, 123, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 133, 121, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 133, 120, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 132, 118, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 132, 117, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 131, 115, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 130, 114, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 130, 113, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 129, 111, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 128, 109, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 128, 107, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 142, 127, 106, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 127, 106, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 127, 104, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 126, 103, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 125, 102, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 125, 102, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 125, 100, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 124, 99, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 124, 99, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 141, 123, 98, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 123, 97, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 122, 95, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 122, 95, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 121, 93, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 121, 93, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 121, 92, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 120, 91, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 120, 90, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 120, 89, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 119, 88, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 118, 87, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 118, 86, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 118, 85, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 117, 85, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 117, 83, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
		{ 140, 116, 82, 0 }, { 132, 127, 127, 0 }, { 146, 141, 141, 0 }, { 162, 156, 156, 0 }, { 176, 170, 170, 0 }, { 189, 183, 183, 0 }, { 199, 192, 192, 0 }, { 209, 202, 202, 0 }, { 217, 209, 209, 0 }, { 224, 216, 216, 0 }, { 228, 221, 221, 0 }, { 231, 223, 223, 0 }, { 233, 225, 225, 0 }, 
	};

	// All 116 colors in the image above.
	RGBQUAD rgbPalette[256] =
	{
		{ 140, 116, 81, 0 }, { 140, 116, 82, 0 }, { 140, 117, 83, 0 }, { 140, 117, 85, 0 },
		{ 140, 118, 85, 0 }, { 140, 118, 86, 0 }, { 140, 118, 87, 0 }, { 140, 119, 88, 0 },
		{ 140, 120, 89, 0 }, { 140, 120, 90, 0 }, { 140, 120, 91, 0 }, { 140, 121, 92, 0 },
		{ 140, 121, 93, 0 }, { 140, 122, 95, 0 }, { 140, 123, 97, 0 }, { 141, 123, 98, 0 },
		{ 141, 124, 99, 0 }, { 141, 125, 100, 0 }, { 141, 125, 102, 0 }, { 141, 126, 103, 0 },
		{ 141, 127, 104, 0 }, { 141, 127, 106, 0 }, { 142, 127, 106, 0 }, { 132, 127, 127, 0 },
		{ 142, 128, 107, 0 }, { 142, 128, 109, 0 }, { 142, 129, 111, 0 }, { 142, 130, 113, 0 },
		{ 142, 130, 114, 0 }, { 142, 131, 115, 0 }, { 142, 132, 117, 0 }, { 142, 132, 118, 0 },
		{ 142, 133, 120, 0 }, { 142, 133, 121, 0 }, { 142, 134, 123, 0 }, { 142, 135, 125, 0 },
		{ 142, 135, 127, 0 }, { 132, 128, 128, 0 }, { 135, 131, 131, 0 }, { 138, 133, 133, 0 },
		{ 142, 136, 128, 0 }, { 142, 137, 129, 0 }, { 142, 138, 131, 0 }, { 142, 138, 132, 0 },
		{ 142, 139, 133, 0 }, { 142, 139, 135, 0 }, { 142, 137, 137, 0 }, { 142, 140, 136, 0 },
		{ 142, 140, 138, 0 }, { 143, 141, 139, 0 }, { 143, 142, 140, 0 }, { 143, 142, 142, 0 },
		{ 146, 141, 141, 0 }, { 147, 142, 142, 0 }, { 149, 144, 144, 0 }, { 151, 146, 146, 0 },
		{ 153, 147, 147, 0 }, { 155, 150, 150, 0 }, { 159, 154, 154, 0 }, { 160, 154, 154, 0 },
		{ 162, 156, 156, 0 }, { 163, 157, 157, 0 }, { 164, 159, 159, 0 }, { 166, 161, 161, 0 },
		{ 168, 162, 162, 0 }, { 169, 163, 163, 0 }, { 170, 164, 164, 0 }, { 173, 167, 167, 0 },
		{ 176, 170, 170, 0 }, { 177, 171, 171, 0 }, { 178, 172, 172, 0 }, { 180, 174, 174, 0 },
		{ 182, 176, 176, 0 }, { 185, 178, 178, 0 }, { 185, 179, 179, 0 }, { 187, 181, 181, 0 },
		{ 189, 183, 183, 0 }, { 190, 184, 184, 0 }, { 192, 185, 185, 0 }, { 194, 187, 187, 0 },
		{ 196, 189, 189, 0 }, { 196, 190, 190, 0 }, { 198, 191, 191, 0 }, { 199, 192, 192, 0 },
		{ 200, 193, 193, 0 }, { 201, 194, 194, 0 }, { 202, 195, 195, 0 }, { 203, 196, 196, 0 },
		{ 205, 198, 198, 0 }, { 207, 200, 200, 0 }, { 209, 202, 202, 0 }, { 210, 203, 203, 0 },
		{ 211, 204, 204, 0 }, { 212, 205, 205, 0 }, { 213, 206, 206, 0 }, { 215, 207, 207, 0 },
		{ 216, 208, 208, 0 }, { 217, 209, 209, 0 }, { 217, 210, 210, 0 }, { 218, 211, 211, 0 },
		{ 219, 212, 212, 0 }, { 220, 213, 213, 0 }, { 221, 214, 214, 0 }, { 222, 214, 214, 0 },
		{ 223, 215, 215, 0 }, { 224, 216, 216, 0 }, { 225, 217, 217, 0 }, { 226, 218, 218, 0 },
		{ 227, 219, 219, 0 }, { 228, 220, 220, 0 }, { 228, 221, 221, 0 }, { 229, 221, 221, 0 },
		{ 230, 222, 222, 0 }, { 231, 223, 223, 0 }, { 232, 224, 224, 0 }, { 233, 225, 225, 0 },
		// rest can be zero.
	};

	const int W = 13;
	const int H = 51;
	const int ColCount = 116;

	BITMAPINFO bmiInput = {0};
	bmiInput.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmiInput.bmiHeader.biWidth = W;
	bmiInput.bmiHeader.biHeight = -H;
	bmiInput.bmiHeader.biPlanes = 1;
	bmiInput.bmiHeader.biBitCount = 32;
	bmiInput.bmiHeader.biCompression = BI_RGB;
	bmiInput.bmiHeader.biSizeImage = 0;
	bmiInput.bmiHeader.biXPelsPerMeter = 0;
	bmiInput.bmiHeader.biYPelsPerMeter = 0;
	bmiInput.bmiHeader.biClrUsed = 0;
	bmiInput.bmiHeader.biClrImportant = 0;

	//BITMAPV5HEADER bmiInput = {0};
	//bmiInput.bV5Size = sizeof(BITMAPV5HEADER);
	//bmiInput.bV5Width = W;
	//bmiInput.bV5Height = -H;
	//bmiInput.bV5Planes = 1;
	//bmiInput.bV5BitCount = 32;
	//bmiInput.bV5Compression = BI_RGB;
	//bmiInput.bV5SizeImage = 0;
	//bmiInput.bV5XPelsPerMeter = 0;
	//bmiInput.bV5YPelsPerMeter = 0;
	//bmiInput.bV5ClrUsed = 0;
	//bmiInput.bV5ClrImportant = 0;
	//bmiInput.bV5CSType = LCS_sRGB; // LCS_WINDOWS_COLOR_SPACE
	//bmiInput.bV5Intent = LCS_GM_ABS_COLORIMETRIC; // LCS_GM_ABS_COLORIMETRIC LCS_GM_BUSINESS LCS_GM_GRAPHICS LCS_GM_IMAGES

	HDC hdcInput = CreateCompatibleDC(0);

	HDC hdcOutput = CreateCompatibleDC(0);

	AutoBuffer< /*BITMAPV5HEADER*/BITMAPINFO > abBmiOutput;
	AutoBuffer< LOGPALETTE > abLogPalette;

	if (hdcInput == NULL)
	{
		MessageBox(hWnd, L"Failed to create hdcInput.", L"Win32UI", MB_OK);
	}
	else if (hdcOutput == NULL)
	{
		MessageBox(hWnd, L"Failed to create hdcOutput.", L"Win32UI", MB_OK);
	}
	else if (!abBmiOutput.AllocateBytes( sizeof(/*BITMAPV5HEADER*/BITMAPINFO) + (sizeof(rgbPalette)/* - sizeof(RGBQUAD)*/) ))
	{
		MessageBox(hWnd, L"Failed to allocate abBmiOutput.", L"Win32UI", MB_OK);
	}
	else if (!abLogPalette.AllocateBytes( sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * 256/*(ColCount - 1)*/ ))
	{
		MessageBox(hWnd, L"Failed to allocate abLogPalette.", L"Win32UI", MB_OK);
	}
	else
	{
		/*
		int icmOrigInput  = SetICMMode(hdcInput,  ICM_QUERY);
		int icmOrigOutput = SetICMMode(hdcOutput, ICM_QUERY);

		switch(icmOrigInput)
		{
		case 0:		             MessageBox(hWnd, L"Failed to query ICM state of input DC (but carrying on anyway).", L"Win32UI", MB_OK); break;
		case ICM_ON:             MessageBox(hWnd, L"Input DC started with ICM on.", L"Win32UI", MB_OK); break;
		case ICM_OFF:            MessageBox(hWnd, L"Input DC started with ICM off.", L"Win32UI", MB_OK); break;
		case ICM_DONE_OUTSIDEDC: MessageBox(hWnd, L"Input DC started with ICM done-outside-dc.", L"Win32UI", MB_OK); break;
		default:                 MessageBox(hWnd, L"Input DC started with ICM in unexpected state.", L"Win32UI", MB_OK); break;
		}

		switch(icmOrigOutput)
		{
		case 0:		             MessageBox(hWnd, L"Failed to query ICM state of output DC (but carrying on anyway).", L"Win32UI", MB_OK); break;
		case ICM_ON:             MessageBox(hWnd, L"Output DC started with ICM on.", L"Win32UI", MB_OK); break;
		case ICM_OFF:            MessageBox(hWnd, L"Output DC started with ICM off.", L"Win32UI", MB_OK); break;
		case ICM_DONE_OUTSIDEDC: MessageBox(hWnd, L"Output DC started with ICM done-outside-dc.", L"Win32UI", MB_OK); break;
		default:                 MessageBox(hWnd, L"Output DC started with ICM in unexpected state.", L"Win32UI", MB_OK); break;
		}
		*/

		/*
		if (0 == SetICMMode(hdcInput, ICM_DONE_OUTSIDEDC))
		{
			MessageBox(hWnd, L"Failed to turn off ICM on input DC (but carrying on anyway).", L"Win32UI", MB_OK);
		}

		if (0 == SetICMMode(hdcOutput, ICM_DONE_OUTSIDEDC))
		{
			MessageBox(hWnd, L"Failed to turn off ICM on output DC (but carrying on anyway).", L"Win32UI", MB_OK);
		}
		*/

		ZeroMemory(abLogPalette.GetBuffer(), abLogPalette.GetSizeBytes());
		abLogPalette.GetBuffer()->palVersion = 0x300;
		abLogPalette.GetBuffer()->palNumEntries = ColCount;

		for (int i = 0; i < ColCount; ++i)
		{
			abLogPalette.GetBuffer()->palPalEntry[i].peRed   = rgbPalette[i].rgbRed;
			abLogPalette.GetBuffer()->palPalEntry[i].peGreen = rgbPalette[i].rgbGreen;
			abLogPalette.GetBuffer()->palPalEntry[i].peBlue  = rgbPalette[i].rgbBlue;
			abLogPalette.GetBuffer()->palPalEntry[i].peFlags = PC_NOCOLLAPSE;
		}

		HPALETTE hPalette = CreatePalette(abLogPalette.GetBuffer());

		if (hPalette == NULL)
		{
			MessageBox(hWnd, L"Failed to create palette.", L"Win32UI", MB_OK);
		}
		else
		{
			void *pInBits = 0;
			HBITMAP hbmInput = CreateDIBSection(hdcInput, (BITMAPINFO*)&bmiInput, DIB_RGB_COLORS, &pInBits, NULL, 0);

			if (hbmInput == NULL)
			{
				MessageBox(hWnd, L"Failed to create hbmInput.", L"Win32UI", MB_OK);
			}
			else
			{
				GdiFlush();
				memcpy(pInBits, rgbInput, sizeof(rgbInput));

				HGDIOBJ hbmOld = SelectObject(hdcInput, hbmInput);

				testImage(hWnd, W, H, hdcInput, rgbInput, L"INPUT");
				testPalette(hWnd, W, H, ColCount, rgbInput, rgbPalette);

				HPALETTE hOldPalette = NULL;
				
				hOldPalette = SelectPalette(hdcOutput, hPalette, FALSE);
				if (hOldPalette == NULL)
				{
					MessageBox(hWnd, L"Failed to select palette (but carrying on anyway).", L"Win32UI", MB_OK);
				}
			//	RealizePalette(hdcOutput); // <-- Doesn't seem to make any difference.

				ZeroMemory(abBmiOutput.GetBuffer(), abBmiOutput.GetSizeBytes());

				memcpy(abBmiOutput.GetBuffer()->bmiColors, rgbPalette, sizeof(rgbPalette));
				abBmiOutput.GetBuffer()->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
				abBmiOutput.GetBuffer()->bmiHeader.biWidth = W;
				abBmiOutput.GetBuffer()->bmiHeader.biHeight = -H;
				abBmiOutput.GetBuffer()->bmiHeader.biPlanes = 1;
				abBmiOutput.GetBuffer()->bmiHeader.biBitCount = 8;
				abBmiOutput.GetBuffer()->bmiHeader.biCompression = BI_RGB;
				abBmiOutput.GetBuffer()->bmiHeader.biSizeImage = 0;
				abBmiOutput.GetBuffer()->bmiHeader.biXPelsPerMeter = 0;
				abBmiOutput.GetBuffer()->bmiHeader.biYPelsPerMeter = 0;
				abBmiOutput.GetBuffer()->bmiHeader.biClrUsed = ColCount;
				abBmiOutput.GetBuffer()->bmiHeader.biClrImportant = ColCount;


				//memcpy(reinterpret_cast< BYTE * >( abBmiOutput.GetBuffer() ) + sizeof(BITMAPV5HEADER), rgbPalette, sizeof(rgbPalette));
				//abBmiOutput.GetBuffer()->bV5Size = sizeof(BITMAPV5HEADER);
				//abBmiOutput.GetBuffer()->bV5Width = W;
				//abBmiOutput.GetBuffer()->bV5Height = -H;
				//abBmiOutput.GetBuffer()->bV5Planes = 1;
				//abBmiOutput.GetBuffer()->bV5BitCount = 8;
				//abBmiOutput.GetBuffer()->bV5Compression = BI_RGB;
				//abBmiOutput.GetBuffer()->bV5SizeImage = 0;
				//abBmiOutput.GetBuffer()->bV5XPelsPerMeter = 0;
				//abBmiOutput.GetBuffer()->bV5YPelsPerMeter = 0;
				//abBmiOutput.GetBuffer()->bV5ClrUsed = ColCount;
				//abBmiOutput.GetBuffer()->bV5ClrImportant = ColCount;
				//abBmiOutput.GetBuffer()->bV5CSType = LCS_sRGB; // LCS_WINDOWS_COLOR_SPACE
				//abBmiOutput.GetBuffer()->bV5Intent = LCS_GM_ABS_COLORIMETRIC; // LCS_GM_ABS_COLORIMETRIC LCS_GM_BUSINESS LCS_GM_GRAPHICS LCS_GM_IMAGES

				void *pOutBits = 0;

				HBITMAP hbmOutput = CreateDIBSection(hdcOutput, (BITMAPINFO*)abBmiOutput.GetBuffer(), DIB_RGB_COLORS, &pOutBits, NULL, 0);

				if (hbmOutput == NULL)
				{
					MessageBox(hWnd, L"Failed to create hbmOutput.", L"Win32UI", MB_OK);
				}
				else
				{
					HGDIOBJ hbmOld2 = SelectObject(hdcOutput, hbmOutput);

					// This should be redundant since the palette was supplied to the DIB on creation, but trying it anyway.
					if (0 == SetDIBColorTable(hdcOutput, 0, ColCount, rgbPalette))
					{
						MessageBox(hWnd, L"SetDIBColorTable failed (but carrying on anyway)", L"Win32UI", MB_OK);
					}

					// I've tried setting the DC's palette here, instead of earlier, but it doesn't seem to change anything.
				//	hOldPalette = SelectPalette(hdcOutput, hPalette, FALSE);
				//	if (hOldPalette == NULL)
				//	{
				//		MessageBox(hWnd, L"Failed to select palette (but carrying on anyway).", L"Win32UI", MB_OK);
				//	}
				//	RealizePalette(hdcOutput);

					COLORREF crWTF = RGB(rgbPalette[0].rgbRed, rgbPalette[0].rgbGreen, rgbPalette[0].rgbBlue);

					if (crWTF != GetNearestColor(hdcOutput, crWTF))
						MessageBox(hWnd, L"Color in palette does not return itself.", L"Win32UI", MB_OK);

					if (0 != GetNearestPaletteIndex(hPalette, crWTF))
						MessageBox(hWnd, L"Color in palette at wrong index.", L"Win32UI", MB_OK);

				//	reportDibColorTable(hWnd, hdcOutput, ColCount);

					if (H != SetDIBitsToDevice(hdcOutput, 0, 0, W, H, 0, 0, 0, H, pInBits, (BITMAPINFO*)&bmiInput, DIB_RGB_COLORS))
					{
						MessageBox(hWnd, L"Failed to set DI bits.", L"Win32UI", MB_OK);
					}
					else
					{
						GdiFlush();
						testImage(hWnd, W, H, hdcOutput, rgbInput, L"OUTPUT SetDIBitsToDevice");
					}

					if (GDI_ERROR == StretchDIBits(hdcOutput, 0, 0, W, H, 0, 0, W, H, pInBits, (BITMAPINFO*)&bmiInput, DIB_RGB_COLORS, SRCCOPY))
					{
						MessageBox(hWnd, L"Failed to StretchDIBits.", L"Win32UI", MB_OK);
					}
					else
					{
						GdiFlush();
						testImage(hWnd, W, H, hdcOutput, rgbInput, L"OUTPUT StretchDIBits");
					}

					if (!BitBlt(hdcOutput, 0, 0, W, H, hdcInput, 0, 0, SRCCOPY))
					{
						MessageBox(hWnd, L"Failed to BitBlt.", L"Win32UI", MB_OK);
					}
					else
					{
						GdiFlush();
						testImage(hWnd, W, H, hdcOutput, rgbInput, L"OUTPUT BitBlt");
					}

					if (!StretchBlt(hdcOutput, 0, 0, W, H, hdcInput, 0, 0, W, H, SRCCOPY))
					{
						MessageBox(hWnd, L"Failed to StretchBlt.", L"Win32UI", MB_OK);
					}
					else
					{
						GdiFlush();
						testImage(hWnd, W, H, hdcOutput, rgbInput, L"OUTPUT StretchBlt");
					}

					for (int y = 0; y < H; ++y)
					{
						for (int x = 0; x < W; ++x)
						{
							SetPixelV(hdcOutput, x, y, GetPixel(hdcInput, x, y));
						}
					}
					GdiFlush();
					testImage(hWnd, W, H, hdcOutput, rgbInput, L"OUTPUT SetPixels");

					SelectObject(hdcOutput, hbmOld2);
					DeleteObject(hbmOutput);
				}

				if (hOldPalette != NULL)
				{
					SelectPalette(hdcOutput, hOldPalette, FALSE);
				}

				SelectObject(hdcInput, hbmOld);
				DeleteObject(hbmInput);
			}

			DeleteObject(hPalette);
		}
	}

	if (hdcOutput)
	{
		DeleteDC(hdcOutput);
	}

	if (hdcInput)
	{
		DeleteDC(hdcInput);
	}

	MessageBox(hWnd, L"End of test 1.", L"Win32UI", MB_OK);
}

void Test2(HWND hWnd)
{
	MessageBox(hWnd, L"Placeholder. (There is no test 2. :))", L"Win32UI", MB_OK);
}
